#
# Copyright 2016 Cumulus Networks, Inc. All rights reserved.
# Author: Daniel Walton, dwalton@cumulusnetworks.com
#
# Portions copyright Gregory Jorgensen 2000-12-24 and licensed under the
# Python Software Foundation license available at https://opensource.org/licenses/Python-2.0
#

import re
import sys
import logging
import sre_constants
from os import listdir

log = logging.getLogger(__name__)

# This is hideous but works
# http://stackoverflow.com/questions/53497/regular-expression-that-matches-valid-ipv6-addresses
# ipv6_regex = '(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))'

# With non-capturing ?: for all of the groups
ipv6_regex = '(?:(?:[0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|(?:[0-9a-fA-F]{1,4}:){1,7}:|(?:[0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|(?:[0-9a-fA-F]{1,4}:){1,5}(?::[0-9a-fA-F]{1,4}){1,2}|(?:[0-9a-fA-F]{1,4}:){1,4}(?::[0-9a-fA-F]{1,4}){1,3}|(?:[0-9a-fA-F]{1,4}:){1,3}(?::[0-9a-fA-F]{1,4}){1,4}|(?:[0-9a-fA-F]{1,4}:){1,2}(?::[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:(?:(?::[0-9a-fA-F]{1,4}){1,6})|:(?:(?::[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(?::[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(?:ffff(?::0{1,4}){0,1}:){0,1}(?:(?:25[0-5]|(?:2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(?:25[0-5]|(?:2[0-4]|1{0,1}[0-9]){0,1}[0-9])|(?:[0-9a-fA-F]{1,4}:){1,4}:(?:(?:25[0-5]|(?:2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(?:25[0-5]|(?:2[0-4]|1{0,1}[0-9]){0,1}[0-9]))'

ipv6_prefixlength_regex = '(?:\d|[1-9]\d|1[0-1]\d|12[0-8])'

# Not as bad as the v6 regex but this does validate that all 4 octets
# are between 0 and 255 (inclusive)
ipv4_regex = '(?:(?:(?:\d|[1-9]\d|1\d{2}|2[0-4]\d|25[0-5])\.){3}(?:\d|[1-9]\d|1\d{2}|2[0-4]\d|25[0-5]))'
ipv4_prefixlength_regex = '(?:\d|[1-2]\d|3[0-2])'

regex_for_range_cache = {}
soundex_cache = {}


def sort_for_humans(list_foo):
    """
    Sort list_foo in the way that humans expect
    http://nedbatchelder.com/blog/200712/human_sorting.html
    """
    convert = lambda text: int(text) if text.isdigit() else text
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]
    list_foo.sort(key=alphanum_key)
    return list_foo


def soundex(name, length=12):
    """
    http://code.activestate.com/recipes/52213-soundex-algorithm/

    soundex module conforming to Knuth's algorithm
    implementation 2000-12-24 by Gregory Jorgensen
    public domain
    """
    global soundex_cache

    # cache the results to save a few cycles
    if name in soundex_cache:
        return soundex_cache[name]

    # digits holds the soundex values for the alphabet
    digits = '01230120022455012623010202'
    sndx = ''
    fc = ''

    # translate alpha chars in name to soundex digits
    for c in name.upper():
        if c.isalpha():
            if not fc:
                fc = c   # remember first letter

            d = digits[ord(c) - ord('A')]

            # duplicate consecutive soundex digits are skipped
            if not sndx or (d != sndx[-1]):
                sndx += d

    # replace first digit with first alpha character
    sndx = fc + sndx[1:]

    # remove all 0s from the soundex code
    sndx = sndx.replace('0', '')

    # pad soundex code to 'length' characters
    sndx = (sndx + (length * '0'))[:length]
    soundex_cache[name] = sndx

    return sndx


# =============================================================================
# https://github.com/dimka665/range-regex
#
#
# Copyright (c) 2013, Dmitry Voronin
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


def _split_to_patterns(min_, max_):
    subpatterns = []

    start = min_
    for stop in _split_to_ranges(min_, max_):
        subpatterns.append(_range_to_pattern(start, stop))
        start = stop + 1

    return subpatterns


def _split_to_ranges(min_, max_):
    stops = {max_}

    nines_count = 1
    stop = _fill_by_nines(min_, nines_count)
    while min_ <= stop < max_:
        stops.add(stop)

        nines_count += 1
        stop = _fill_by_nines(min_, nines_count)

    zeros_count = 1
    stop = _fill_by_zeros(max_ + 1, zeros_count) - 1
    while min_ < stop <= max_:
        stops.add(stop)

        zeros_count += 1
        stop = _fill_by_zeros(max_ + 1, zeros_count) - 1

    stops = list(stops)
    stops.sort()

    return stops


def _fill_by_nines(integer, nines_count):
    return int(str(integer)[:-nines_count] + '9' * nines_count)


def _fill_by_zeros(integer, zeros_count):
    return integer - integer % 10 ** zeros_count


def _range_to_pattern(start, stop):
    pattern = ''
    any_digit_count = 0

    for start_digit, stop_digit in zip(str(start), str(stop)):
        if start_digit == stop_digit:
            pattern += start_digit
        elif start_digit != '0' or stop_digit != '9':
            pattern += '[{}-{}]'.format(start_digit, stop_digit)
        else:
            any_digit_count += 1

    if any_digit_count:
        pattern += r'\d'

    if any_digit_count > 1:
        pattern += '{{{}}}'.format(any_digit_count)

    return pattern


def regex_for_range(min_, max_):
    """
    > regex_for_range(12, 345)
    '1[2-9]|[2-9]\d|[1-2]\d{2}|3[0-3]\d|34[0-5]'
    """
    positive_subpatterns = []
    negative_subpatterns = []

    if min_ < 0:
        min__ = 1
        if max_ < 0:
            min__ = abs(max_)
        max__ = abs(min_)

        negative_subpatterns = _split_to_patterns(min__, max__)
        min_ = 0

    if max_ >= 0:
        positive_subpatterns = _split_to_patterns(min_, max_)

    negative_only_subpatterns = ['-' + val for val in negative_subpatterns if val not in positive_subpatterns]
    positive_only_subpatterns = [val for val in positive_subpatterns if val not in negative_subpatterns]
    intersected_subpatterns = ['-?' + val for val in negative_subpatterns if val in positive_subpatterns]

    subpatterns = negative_only_subpatterns + intersected_subpatterns + positive_only_subpatterns
    return '|'.join(subpatterns)

# End of range-regex
# ===========================================


class Token():

    def __init__(self, index, text):

        # Remove the leading/trailing () or []
        if text.startswith('('):
            self.text = text[1:-1]
            self.optional = False
        elif text.startswith('['):
            self.text = text[1:-1]
            self.optional = True
        elif text.startswith('-'):
            self.text = text
            self.optional = True
        else:
            self.text = text
            self.optional = False

        self.index = index
        self.words = self.text.split('|')
        self.matched_word = None
        self.value = False
        self.exact_match = False

        # Build a list of regular expressions to match each word in this token
        # So if self.words are ['add', 'de]' then  self.regex_per_word will be ['(add|ad|a)', '(del|de|d)']
        self.regex_per_word = []

        for word in self.words:
            if word.startswith('<'):
                if word == '<ip>' or word == '<source-ip>':
                    self.regex_per_word.append("%s|%s" % (ipv4_regex, ipv6_regex))

                elif word == '<ipv4>':
                    self.regex_per_word.append(ipv4_regex)

                elif word == '<ipv4/prefixlen>':
                    self.regex_per_word.append('%s\/%s' % (ipv4_regex, ipv4_prefixlength_regex))

                elif word == '<ipv6>':
                    self.regex_per_word.append(ipv6_regex)

                elif word == '<ipv6/prefixlen>':
                    self.regex_per_word.append('%s\/%s' % (ipv6_regex, ipv6_prefixlength_regex))

                elif word == '<ip/prefixlen>':
                    self.regex_per_word.append('%s\/%s|%s\/%s' %
                                               (ipv4_regex, ipv4_prefixlength_regex, ipv6_regex, ipv6_prefixlength_regex))

                elif word == '<major>' or word == '<minor>' or word == '<number>':
                    self.regex_per_word.append('\d+')

                elif word == '<text>':
                    self.regex_per_word.append('\S+')

                # This needs some more so we can tab complete interface names
                elif word == '<interface>' or word == '<interface-as-arg>' or word == '<iface>':
                    self.regex_per_word.append('\w+\S+')

                elif word == '<number-ipv4-list>':
                    self.regex_per_word.append('(?:\d+=\d+\.\d+\.\d+\.\d+\s?)+')

                elif word == '<wildcard>':
                    self.regex_per_word.append('.*')

                elif word == '<mac>':
                    # ?: makes this a non-capturing group
                    self.regex_per_word.append('(?:[0-9A-Fa-f]{2}[:-]){5}(?:[0-9A-Fa-f]{2})')

                else:
                    re_number_range = re.match('<(\d+)-(\d+)>', word)

                    if re_number_range:
                        min_int = int(re_number_range.group(1))
                        max_int = int(re_number_range.group(2))

                        if (min_int, max_int) not in regex_for_range_cache:
                            regex = regex_for_range(min_int, max_int)
                            regex_for_range_cache[(min_int, max_int)] = regex
                        self.regex_per_word.append(regex_for_range_cache[(min_int, max_int)])
                    else:
                        raise Exception("Add token regex support for %s" % word)

            # Keyword
            else:
                if word == '*':
                    self.regex_per_word.append('\*')
                elif word.startswith('-'):
                    self.regex_per_word.append(word)
                else:
                    # For 'add' we want '(add|ad|a)' so we can match on partial keywords
                    # Do not do this for "1000" though, we do not want (1000|100|10|1)
                    # Also do not do this for "--json" or "-j"
                    keyword_regex = []

                    while word:
                        keyword_regex.append(word)

                        if word.isdigit():
                            break
                        else:
                            word = word[0:-1]

                    self.regex_per_word.append('(?:' + '|'.join(keyword_regex) + ')')

        self.regex_per_word_ambiguous = list(self.regex_per_word)

        # Say our two options are (off|on), our regexes for this pair would be
        # ((off|of|o)|(on|o)) so 'on' would end up matching (off|of|o) due to the 'o'
        #
        # For these scenarios we must remove the ambiguous matching options.
        # Build a dictionary so we can count how many times each sub-token is used.
        # So for 'o' in our (off|on) example the sub_tokens dictionary would be {'o' : 2}
        sub_tokens = {}
        for (i, word) in enumerate(self.words):

            # Keyword
            if not word.startswith('<') and not word.startswith('-'):
                # 3:-1 to remove the leading (?: and the trailing ), then split by |
                for sub_token in self.regex_per_word[i][3:-1].split('|'):
                    if sub_token != word:
                        if sub_token not in sub_tokens:
                            sub_tokens[sub_token] = 0
                        sub_tokens[sub_token] += 1

        # Create a list of the sub-tokens that need to be removed
        sub_tokens_to_del = [sub_token for (sub_token, count) in sub_tokens.iteritems() if count > 1]

        # Rebuild each self.regex_per_word and remove any sub-token in sub_tokens_to_del
        if sub_tokens_to_del:
            for (i, word) in enumerate(self.words):

                # Keyword
                if not word.startswith('<') and not word.startswith('-'):
                    keyword_regex = []

                    # 3:-1 to remove the leading (?: and the trailing ), then split by |
                    for sub_token in self.regex_per_word[i][3:-1].split('|'):
                        if sub_token not in sub_tokens_to_del:
                            keyword_regex.append(sub_token)

                    self.regex_per_word[i] = '(?:' + '|'.join(keyword_regex) + ')'

        self.regex = '(' + '|'.join(self.regex_per_word) + ')'
        self.regex_ambiguous = '(' + '|'.join(self.regex_per_word_ambiguous) + ')'

        # ? means match 0 or 1 of everything in the preceding ()
        if self.optional:
            self.regex += '?'

    def __str__(self):
        return "WORDS: %-40s, MATCHED WORD: %-20s, VALUE: %s" % \
            (self.words, self.matched_word, self.value)


class CommandSequence():

    def __init__(self, string, interfaces):
        self.options = []
        self.interfaces = interfaces

        # Remove leading and trailing whitespaces
        self.text = string.strip()

        # Compress duplicate spaces into single space
        self.text = ' '.join(self.text.split()[1:])

        # Remove spaces we do not need
        self.text = self.text.replace(' |', '|')
        self.text = self.text.replace('| ', '|')
        self.text = self.text.replace(' ]', ']')
        self.text = self.text.replace('[ ', '[')

        assert self.text.count('[') == self.text.count(']'), 'You have mis-matched []s in:\n  %s\n' % self.text
        assert self.text.count('(') == self.text.count(')'), 'You have mis-matched ()s in:\n  %s\n' % self.text

        # TODO - this breaks for "bgp (<ipv4>|<ipv6>|<interface>|*) [soft in|soft out]"
        # Now that we know each word is separated by a single whitespace split
        # on whitespace. Ignore the first token which is the program name.
        self.tokens = []
        for (index, text) in enumerate(self.text.split()):
            self.tokens.append(Token(index, text))

        self.regex = self.get_regex_for_tokens(self.tokens)
        self.len_tokens = len(self.tokens)

    def __str__(self):
        token_text = []
        for (index, token) in enumerate(self.tokens):
            token_text.append('%d: %s' % (index, token))
        return "%s\n%s\nREGEX: %s\nOPTIONS: %s\n" % (self.text, '\n'.join(token_text), self.regex, self.options)

    def get_regex_for_tokens(self, tokens, use_ambiguous=False, ends_with_space=False):
        regex_words = ['^', ]
        prev_token = None

        for token in tokens:
            if prev_token:
                if token.optional or prev_token.optional:
                    regex_words.append('\s?')
                else:
                    regex_words.append(' ')

            regex_words.append(token.regex)
            prev_token = token

        # For this command:
        # ip ospf network (point-to-multipoint|point-to-point)
        #
        # "ip ospf network po<TAB>" needs to return point-to-multipoint and point-to-point
        # Our normal regex for this token won't work though because it looks like:
        #  ^((?:add|ad|a)|(?:del|de|d)) (\w+\S+) ((?:ip|i)) ((?:ospf|osp|os|o)) ((?:network|networ|netwo|netw|net|ne|n)) ((?:broadcast|broadcas|broadca|broadc|broad|broa|bro|br|b)|(?:non-broadcast|non-broadcas|non-broadca|non-broadc|non-broad|non-broa|non-bro|non-br|non-b|non-|non|no|n)|(?:point-to-multipoint|point-to-multipoin|point-to-multipoi|point-to-multipo|point-to-multip|point-to-multi|point-to-mult|point-to-mul|point-to-mu|point-to-m)|(?:point-to-point|point-to-poin|point-to-poi|point-to-po|point-to-p))\s$
        #
        # We must use an ambiguous regex here
        #  ^((?:add|ad|a)|(?:del|de|d)) (\w+\S+) ((?:ip|i)) ((?:ospf|osp|os|o)) ((?:network|networ|netwo|netw|net|ne|n)) ((?:broadcast|broadcas|broadca|broadc|broad|broa|bro|br|b)|(?:non-broadcast|non-broadcas|non-broadca|non-broadc|non-broad|non-broa|non-bro|non-br|non-b|non-|non|no|n)|(?:point-to-multipoint|point-to-multipoin|point-to-multipoi|point-to-multipo|point-to-multip|point-to-multi|point-to-mult|point-to-mul|point-to-mu|point-to-m|point-to-|point-to|point-t|point-|point|poin|poi|po|p)|(?:point-to-point|point-to-poin|point-to-poi|point-to-po|point-to-p|point-to-|point-to|point-t|point-|point|poin|poi|po|p))\s$
        #
        # TODO: this needs some more work
        # if use_ambiguous and prev_token:
        #     regex_words = regex_words[:-1]
        #     regex_words.append(prev_token.regex_ambiguous)

        if ends_with_space:
            regex_words.append('\s')

        regex_words.append('$')
        return ''.join(regex_words)


    def get_soundex_score(self, argv, len_argv, ended_with_space=False):
        """
        Loop through all of the tokens and look for the ones that are standalone keywords.
        So from "<interface> (set|clear) mstpctl-portpathcost <0-65535>" we will only
        examine the tokens for set, clear, and mstpctl-portpathcost.

        For each of those compare their soundex value vs. the soundex value of the words
        entered by the user on the command line
        """
        soundex_score = 0

        for argv_word in argv:
            soundex_argv_word = soundex(argv_word)
            matched = False

            for token in self.tokens:
                for token_word in token.words:

                    # Do not compute soundex for *, <ipv4>, etc
                    if token_word == '*' or token_word.startswith('<'):
                        continue

                    soundex_token_word = soundex(token_word)

                    if soundex_argv_word == soundex_token_word:
                        log.debug("SOUNDEX MATCH: %s vs. %s" % (token_word, argv_word))
                        soundex_score += 1
                        matched = True
                        break

                # If one of the words in this token matched then break out of this
                # loop to avoid a possible double count for this argv_word
                if matched:
                    break

        return soundex_score

    def words_for_regex(self):
        """
        Return the min and max number of words that can be used to match this command.
        Also return a list of required starting letters that must be present for this
        regex to match.  For example if the command is
            quagga show bgp [ipv4|ipv6] [unicast|multicast] summary [json]

        then 's' and 'b' are the required starting letters.
        """
        min_words = 0
        max_words = 0
        required_starting_letters = []

        for word in self.text.split():
            max_words += 1

            if not word.startswith('['):
                min_words += 1

                # Is the first character of the word a letter? (ie not *, <, etc)
                if word[0].isalpha():
                    required_starting_letters.append(word[0])

        required_starting_letters = list(set(required_starting_letters))

        # log.debug("TEXT : %s" % self.text)
        # log.debug("min_words %d, max_words %d, required_starting_letters %s" %\
        #           (min_words, max_words, required_starting_letters))
        return (min_words, max_words, required_starting_letters)

    def argv_has_starting_letters(self, argv, required_starting_letters):

        for letter in required_starting_letters:
            for word in argv:
                if word.startswith(letter):
                    break
            else:
                return False

        return True

    def argv_matches_regex(self, argv, len_argv, ended_with_space=False):
        argv_string = ' '.join(argv)
        self.options = []

        # Calling re.match hundreds of times can be expensive/slow...do some basic
        # sanity checking to see if we can return False without doing the full
        # regex evaluation. For programs with 100+ cmds this can cut the time to
        # evaluate all of the commands by a factor of 10x
        (min_len_argv, max_len_argv, required_starting_letters) = self.words_for_regex()

        if len_argv < min_len_argv:
            log.debug('\033[94mSKIP (%d < min-length %d)\033[0m: %s' % (len_argv, min_len_argv, self.text))
            return False

        if len_argv > max_len_argv and '<wildcard>' not in self.text:
            log.debug('\033[94mSKIP (%d > max-length %d)\033[0m: %s' % (len_argv, max_len_argv, self.text))
            return False

        if not self.argv_has_starting_letters(argv, required_starting_letters):
            log.debug('\033[94mSKIP (required letters %s)\033[0m: %s' % (required_starting_letters, self.text))
            return False

        # Evaluate the command line string vs. the regex for this CommandSequence
        try:
            re_match = re.match(self.regex, argv_string)
        except sre_constants.error as e:
            log.error("Invalid regex for '%s':\n\n%s\n" % (self.text, self.regex))
            raise e

        if re_match:
            log.debug('')
            log.debug('\033[92mDID MATCH\033[0m: %s' % self.text)
            log.debug("ARGV_STRING: '%s'" % argv_string)
            log.debug("REGEX: %s" % self.regex)
            log.debug("RE_GROUPS : %s" % str(re_match.groups()))
            self.last_matching_token = None

            for (i, value) in enumerate(re_match.groups()):
                token = self.tokens[i]

                if value is None:
                    continue

                # Evaluate value vs the regex for each word in the token.  So if the
                # user entered 'ad' and the choices where (add|del) we need to figure
                # out that this matched against 'add' and not 'del'.  We do this by
                # attempting to match 'ad' against regex_per_word which are (add|ad|a)
                # and (del|de|d) in this case.
                log.debug("TOKEN WORDS[%d]: %s -> %s" % (i + 1, token.words, value))

                for (j, word) in enumerate(token.words):

                    if re.match(token.regex_per_word[j], value):
                        log.debug("TOKEN REGEX DID MATCH (per word): %s" % token.regex_per_word[j])

                        # value will be the string entered on the command line. So if they
                        # entered 1.1.1.1/32 for <ip/prefixlen> then token.matched_word will
                        # be '<ip/prefixlen>' and token.value will be 1.1.1.1/32
                        token.matched_word = word

                        # Determine if this was an exact match
                        if token.matched_word.startswith('<'):
                            if token.matched_word.startswith('<interface') or token.matched_word.startswith('<iface'):
                                if value in self.interfaces:
                                    token.exact_match = True
                            else:
                                token.exact_match = True

                        # keyword
                        else:
                            if value == token.words[j]:
                                token.exact_match = True

                        # If this is a <> then the value is the input the user entered
                        # If not then value is True
                        if token.matched_word.startswith('<'):
                            if value.isdigit():
                                token.value = int(value)
                            else:
                                token.value = value

                        # keyword...use True for the value
                        else:
                            token.value = True

                        self.last_matching_token = token
                        break
                else:
                    log.debug("TOKEN REGEX DID NOT MATCH: %s" % token.regex)
                    for (j, word) in enumerate(token.words):
                        log.debug("TOKEN REGEX (per word): %s" % token.regex_per_word[j])
                    raise Exception("Could not find the token that we matched")


            if self.last_matching_token:
                if self.last_matching_token.exact_match and ended_with_space:
                    if self.last_matching_token != self.tokens[-1]:
                        self.options = self.tokens[self.last_matching_token.index + 1].words
                else:
                    if (self.last_matching_token.matched_word.startswith('<interface') or
                        self.last_matching_token.matched_word.startswith('<iface')) :

                        # If the user entered 'swp' as an interface name then only list
                        # the interfaces that begin with 'swp', not 'lo', 'br', etc
                        for ifname in self.interfaces:
                            if ifname.startswith(self.last_matching_token.value):
                                self.options.append(ifname)
                    else:
                        self.options = [self.last_matching_token.matched_word, ]
            else:
                self.options = self.tokens[0].words

            log.debug("OPTIONS: %s\n" % self.options)
            return True

        # re_match was false
        else:
            log.debug("\033[91mDID NOT MATCH\033[0m: %s" % self.text)
            # log.debug("ARGV_STRING: '%s'" % argv_string)
            # log.debug("REGEX: %s" % self.regex)
            # log.debug('')
            return False


class NetworkDocopt():
    """
    Heavily influenced by docopt but designed to be a little more like
    a Networking CLI with partial word acceptance, IPv4 sanity checking, etc
    """

    def __init__(self, docstring):
        self.args = {}
        self.match = None
        self.options = []
        self.program = None
        self.docstring = docstring
        self.commands = []
        interfaces = [x for x in listdir('/sys/class/net/') if x != 'bonding_masters']

        # The 1st item in argv is the program name...ignore it
        self.argv = sys.argv[1:]

        # Parse the 'Usage' section of the doc string.  Create a
        # CommandSequence object for every line in Usage, store those
        # objects on the self.commands list.
        state = None
        found_help_option = False

        for line in docstring.split('\n'):

            if line.startswith('Usage'):
                state = 'Usage'
                continue
            elif line.startswith('Help'):
                break
            elif line.startswith('Options'):
                break
            # Ignore comments in the docstring
            elif line.lstrip().startswith('#'):
                continue
            elif line == '':
                continue

            if state == 'Usage':
                if '--help' in line:
                    found_help_option = True

                self.commands.append(CommandSequence(line, interfaces))

                if self.program is None:
                    result = re.match('^\s+(\S+)', line)
                    if result:
                        self.program = result.group(1)

        # Automagically add a (-h|--help) option
        if not found_help_option:
            line = '%s (-h|--help)' % self.program
            self.commands.append(CommandSequence(line, interfaces))

        if logging.getLogger().getEffectiveLevel() == logging.DEBUG:
            for cmd in self.commands:
                log.debug('CommandSequence: %s' % cmd)

    def get(self, keyword):
        return self.args.get(keyword)

    def run(self):

        # Loop over all of the CommandSequence objects and build a list
        # of every kind of token in the doc string
        all_tokens = []
        for cmd in self.commands:
            for token in cmd.tokens:
                all_tokens += token.words
        all_tokens = set(all_tokens)

        # Init all tokens in args to None
        for x in all_tokens:
            self.args[x] = None

        log.debug("ARGS STRING: '%s'" % ' '.join(self.argv))
        candidates = []
        len_argv = len(self.argv)

        for cmd in self.commands:
            if cmd.argv_matches_regex(self.argv, len_argv):
                candidates.append(cmd)

        if not candidates:
            print 'No match found for "%s"' % ' '.join(self.argv)
            did_you_mean = []
            program_name = sys.argv[0]
            if '/' in program_name:
                program_name = program_name.split('/')[-1]

            for cmd in self.commands:
                log.debug('SOUNDEX QUERY: %s' % cmd.text)
                soundex_score = cmd.get_soundex_score(self.argv, len_argv)

                if soundex_score:
                    did_you_mean.append((soundex_score, program_name + ' ' + cmd.text))
                log.debug('')

            if did_you_mean:
                prev_score = None
                did_you_mean_highest_score = []

                # Sort these from highest score to lowest, only print the entries
                # that are tied for the highest score
                for (score, text) in reversed(sorted(did_you_mean)):
                    if prev_score is None:
                        did_you_mean_highest_score.append(text)
                    else:
                        if score == prev_score:
                            did_you_mean_highest_score.append(text)
                        else:
                            break
                    prev_score = score

                did_you_mean_highest_score = sorted(did_you_mean_highest_score)
                print 'Did you mean one of the following?\n\n    %s' % '\n    '.join(did_you_mean_highest_score)

            # We have no clue as to what set of options the user was trying
            # to enter, print the entire docstring.  The user has to enter some
            # fairly off the wall input for us to get to this point.
            else:
                print(self.docstring)

            sys.exit(1)

        elif len(candidates) == 1:
            log.debug("There is one candidate")
            cmd = candidates[0]
            log.debug(cmd)

            for token in cmd.tokens:
                self.args[token.matched_word] = token.value
                log.debug("args key: %s, value: %s" % (token.matched_word, token.value))

            self.match = cmd

            # If the user entered -h or --help print the docstring and exit
            if cmd.len_tokens == 1:
                token = cmd.tokens[0]

                if token.matched_word == '-h' or token.matched_word == '--help':
                    print(self.docstring)
                    sys.exit(0)
        else:
            print "Ambiguous Match. Options:\n\n%s\n" % '\n'.join([cmd.text for cmd in candidates])
            sys.exit(1)

    def print_options(self, ended_with_space):

        # Find the max tokens among all commands
        max_tokens = 0
        for cmd in self.commands:
            if cmd.len_tokens > max_tokens:
                max_tokens = cmd.len_tokens

        no_tokens_chopped_options = []
        tokens_chopped_options = []

        len_argv = len(self.argv)

        if ended_with_space:
            self.argv.append('')

        # We need to find the the options
        # To do this we start with the longest command and start chopping off tokens until
        # - the regex of the remaining tokens matches
        # - we run out of tokens
        for target_token_length in reversed(range(0, max_tokens + 1)):
            log.debug("Removing all but %d tokens from each cmd" % target_token_length)

            for cmd in self.commands:
                if cmd.len_tokens >= target_token_length:
                    tokens_to_chop = cmd.len_tokens - target_token_length

                    # If ended_with_space is true then we almost always want to
                    # add a \s to the regex. The exception is if the user did
                    # "netshow <TAB>" so that they can see all of the first level
                    # options. In that scenario we want a ^$ regex, not ^\s$.
                    ends_with_space = ended_with_space and cmd.len_tokens > tokens_to_chop

                    if tokens_to_chop:
                        cmd.regex = cmd.get_regex_for_tokens(cmd.tokens[:-1 * tokens_to_chop], True, ends_with_space)
                        # remove the last word from cmd.text
                        cmd.text = ' '.join(cmd.text.split()[:-1])
                    else:
                        cmd.regex = cmd.get_regex_for_tokens(cmd.tokens, True, ends_with_space)

                    if cmd.argv_matches_regex(self.argv, len_argv, ended_with_space):
                        if tokens_to_chop:
                            tokens_chopped_options.extend(cmd.options)
                        else:
                            no_tokens_chopped_options.extend(cmd.options)
            log.debug("")

        # In a previous life we needed to distinguish between the options available
        # when a token was chopped vs. those that were available when no tokens were chopped.
        # that is no longer the case today so just combine the two lists.
        log.debug("no_tokens_chopped_options: %s" % str(no_tokens_chopped_options))
        log.debug("tokens_chopped_options: %s" % str(tokens_chopped_options))
        self.options = no_tokens_chopped_options + tokens_chopped_options

        if self.options:
            self.options = sort_for_humans(list(set(self.options)))
            print('\n'.join(self.options))

            # If the only option is something like "<ip>" we cannot print just "<ip>"
            # because bash will treat this as a pure string...it doesn't know that we
            # do not mean the characters "< i p >".  Print '' too, just so there is more
            # than one option.
            if len(self.options) == 1 and self.options[0].startswith('<'):
                print "''"
        else:
            log.debug('There are no options')
